@@ -1,20 +1,22 @@
 """A class to manage fetching and caching of images by URL"""
-
+#
 # Copyright 2012-2014 Anthony Beville
-
+#
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
 # You may obtain a copy of the License at
-
+#
 #     http://www.apache.org/licenses/LICENSE-2.0
-
+#
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS,
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+from __future__ import annotations
 
 import datetime
+import logging
 import os
 import shutil
 import sqlite3 as lite
@@ -22,6 +24,9 @@ import tempfile
 
 import requests
 
+from comictaggerlib import ctversion
+from comictaggerlib.settings import ComicTaggerSettings
+
 try:
     from PyQt5 import QtCore, QtNetwork
 
@@ -29,20 +34,14 @@ try:
 except ImportError:
     qt_available = False
 
-
-import logging
-
-from comictaggerlib import ctversion
-from comictaggerlib.settings import ComicTaggerSettings
-
 logger = logging.getLogger(__name__)
 
 
 class ImageFetcherException(Exception):
-    pass
+    ...
 
 
-def fetch_complete(this, image_data):
+def fetch_complete(image_data: bytes | QtCore.QByteArray) -> None:
     ...
 
 
@@ -50,7 +49,7 @@ class ImageFetcher:
 
     image_fetch_complete = fetch_complete
 
-    def __init__(self):
+    def __init__(self) -> None:
 
         self.settings_folder = ComicTaggerSettings.get_settings_folder()
         self.db_file = os.path.join(self.settings_folder, "image_url_cache.db")
@@ -65,12 +64,12 @@ class ImageFetcher:
         if qt_available:
             self.nam = QtNetwork.QNetworkAccessManager()
 
-    def clear_cache(self):
+    def clear_cache(self) -> None:
         os.unlink(self.db_file)
         if os.path.isdir(self.cache_folder):
             shutil.rmtree(self.cache_folder)
 
-    def fetch(self, url, blocking=False):
+    def fetch(self, url: str, blocking: bool = False) -> bytes:
         """
         If called with blocking=True, this will block until the image is
         fetched.
@@ -83,7 +82,7 @@ class ImageFetcher:
         # first look in the DB
         image_data = self.get_image_from_cache(url)
         if blocking or not qt_available:
-            if image_data is None:
+            if not image_data:
                 try:
                     image_data = requests.get(url, headers={"user-agent": "comictagger/" + ctversion.version}).content
                 except Exception as e:
@@ -96,8 +95,8 @@ class ImageFetcher:
 
         if qt_available:
             # if we found it, just emit the signal asap
-            if image_data is not None:
-                self.image_fetch_complete(QtCore.QByteArray(image_data))
+            if image_data:
+                ImageFetcher.image_fetch_complete(QtCore.QByteArray(image_data))
                 return bytes()
 
             # didn't find it.  look online
@@ -107,7 +106,7 @@ class ImageFetcher:
             # we'll get called back when done...
         return bytes()
 
-    def finish_request(self, reply):
+    def finish_request(self, reply: QtNetwork.QNetworkReply) -> None:
         # read in the image data
         logger.debug("request finished")
         image_data = reply.readAll()
@@ -115,9 +114,9 @@ class ImageFetcher:
         # save the image to the cache
         self.add_image_to_cache(self.fetched_url, image_data)
 
-        self.image_fetch_complete(image_data)
+        ImageFetcher.image_fetch_complete(image_data)
 
-    def create_image_db(self):
+    def create_image_db(self) -> None:
 
         # this will wipe out any existing version
         open(self.db_file, "wb").close()
@@ -135,7 +134,7 @@ class ImageFetcher:
 
             cur.execute("CREATE TABLE Images(url TEXT,filename TEXT,timestamp TEXT,PRIMARY KEY (url))")
 
-    def add_image_to_cache(self, url, image_data):
+    def add_image_to_cache(self, url: str, image_data: bytes | QtCore.QByteArray) -> None:
 
         con = lite.connect(self.db_file)
 
@@ -145,13 +144,12 @@ class ImageFetcher:
             timestamp = datetime.datetime.now()
 
             tmp_fd, filename = tempfile.mkstemp(dir=self.cache_folder, prefix="img")
-            f = os.fdopen(tmp_fd, "w+b")
-            f.write(image_data)
-            f.close()
+            with os.fdopen(tmp_fd, "w+b") as f:
+                f.write(bytes(image_data))
 
             cur.execute("INSERT or REPLACE INTO Images VALUES(?, ?, ?)", (url, filename, timestamp))
 
-    def get_image_from_cache(self, url):
+    def get_image_from_cache(self, url: str) -> bytes:
 
         con = lite.connect(self.db_file)
         with con:
@@ -161,16 +159,16 @@ class ImageFetcher:
             row = cur.fetchone()
 
             if row is None:
-                return None
+                return bytes()
 
             filename = row[0]
-            image_data = None
+            image_data = bytes()
 
             try:
                 with open(filename, "rb") as f:
                     image_data = f.read()
                     f.close()
-            except IOError:
+            except OSError:
                 pass
 
             return image_data
