@@ -1,20 +1,21 @@
 """A python app to (automatically) tag comic archives"""
-
+#
 # Copyright 2012-2014 Anthony Beville
-
+#
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
 # You may obtain a copy of the License at
-
+#
 #     http://www.apache.org/licenses/LICENSE-2.0
-
+#
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS,
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+from __future__ import annotations
 
-import logging
+import json
 import logging.handlers
 import os
 import pathlib
@@ -22,31 +23,37 @@ import platform
 import signal
 import sys
 import traceback
+import types
 
-import pkg_resources
-
+from comicapi import utils
 from comictaggerlib import cli
 from comictaggerlib.comicvinetalker import ComicVineTalker
 from comictaggerlib.ctversion import version
-from comictaggerlib.options import Options
+from comictaggerlib.options import parse_cmd_line
 from comictaggerlib.settings import ComicTaggerSettings
 
+if sys.version_info < (3, 10):
+    import importlib_metadata
+else:
+    import importlib.metadata as importlib_metadata
+
 logger = logging.getLogger("comictagger")
 logging.getLogger("comicapi").setLevel(logging.DEBUG)
+logging.getLogger("comictaggerlib").setLevel(logging.DEBUG)
 logger.setLevel(logging.DEBUG)
 
 try:
     qt_available = True
     from PyQt5 import QtCore, QtGui, QtWidgets
 
-    def show_exception_box(log_msg):
+    def show_exception_box(log_msg: str) -> None:
         """Checks if a QApplication instance is available and shows a messagebox with the exception message.
         If unavailable (non-console application), log an additional notice.
         """
         if QtWidgets.QApplication.instance() is not None:
             errorbox = QtWidgets.QMessageBox()
-            errorbox.setText("Oops. An unexpected error occured:\n{0}".format(log_msg))
-            errorbox.exec_()
+            errorbox.setText(f"Oops. An unexpected error occurred:\n{log_msg}")
+            errorbox.exec()
             QtWidgets.QApplication.exit(1)
         else:
             logger.debug("No QApplication instance available.")
@@ -54,8 +61,8 @@ try:
     class UncaughtHook(QtCore.QObject):
         _exception_caught = QtCore.pyqtSignal(object)
 
-        def __init__(self, *args, **kwargs):
-            super(UncaughtHook, self).__init__(*args, **kwargs)
+        def __init__(self) -> None:
+            super().__init__()
 
             # this registers the exception_hook() function as hook with the Python interpreter
             sys.excepthook = self.exception_hook
@@ -63,7 +70,9 @@ try:
             # connect signal to execute the message box function always on main thread
             self._exception_caught.connect(show_exception_box)
 
-        def exception_hook(self, exc_type, exc_value, exc_traceback):
+        def exception_hook(
+            self, exc_type: type[BaseException], exc_value: BaseException, exc_traceback: types.TracebackType | None
+        ) -> None:
             """Function handling uncaught exceptions.
             It is triggered each time an uncaught exception occurs.
             """
@@ -72,12 +81,8 @@ try:
                 sys.__excepthook__(exc_type, exc_value, exc_traceback)
             else:
                 exc_info = (exc_type, exc_value, exc_traceback)
-                log_msg = "\n".join(
-                    ["".join(traceback.format_tb(exc_traceback)), "{0}: {1}".format(exc_type.__name__, exc_value)]
-                )
-                logger.critical(
-                    "Uncaught exception: %s", "{0}: {1}".format(exc_type.__name__, exc_value), exc_info=exc_info
-                )
+                log_msg = "\n".join(["".join(traceback.format_tb(exc_traceback)), f"{exc_type.__name__}: {exc_value}"])
+                logger.critical("Uncaught exception: %s: %s", exc_type.__name__, exc_value, exc_info=exc_info)
 
                 # trigger message box show
                 self._exception_caught.emit(log_msg)
@@ -85,19 +90,32 @@ try:
     qt_exception_hook = UncaughtHook()
     from comictaggerlib.taggerwindow import TaggerWindow
 except ImportError as e:
+
+    def show_exception_box(log_msg: str) -> None:
+        ...
+
     logger.error(str(e))
     qt_available = False
 
 
-def rotate(handler: logging.handlers.RotatingFileHandler, filename: pathlib.Path):
+def rotate(handler: logging.handlers.RotatingFileHandler, filename: pathlib.Path) -> None:
     if filename.is_file() and filename.stat().st_size > 0:
         handler.doRollover()
 
 
-def ctmain():
-    opts = Options()
-    opts.parse_cmd_line_args()
-    SETTINGS = ComicTaggerSettings(opts.config_path)
+def update_publishers() -> None:
+    json_file = ComicTaggerSettings.get_settings_folder() / "publishers.json"
+    if json_file.exists():
+        try:
+            utils.update_publishers(json.loads(json_file.read_text("utf-8")))
+        except Exception as e:
+            logger.exception("Failed to load publishers from %s", json_file)
+            show_exception_box(str(e))
+
+
+def ctmain() -> None:
+    opts = parse_cmd_line()
+    settings = ComicTaggerSettings(opts.config_path)
 
     os.makedirs(ComicTaggerSettings.get_settings_folder() / "logs", exist_ok=True)
     stream_handler = logging.StreamHandler()
@@ -118,15 +136,17 @@ def ctmain():
     # Need to load setting before anything else
 
     # manage the CV API key
-    if opts.cv_api_key:
-        if opts.cv_api_key != SETTINGS.cv_api_key:
-            SETTINGS.cv_api_key = opts.cv_api_key
-            SETTINGS.save()
-    if opts.only_set_key:
-        print("Key set")
+    # None comparison is used so that the empty string can unset the value
+    if opts.cv_api_key is not None or opts.cv_url is not None:
+        settings.cv_api_key = opts.cv_api_key if opts.cv_api_key is not None else settings.cv_api_key
+        settings.cv_url = opts.cv_url if opts.cv_url is not None else settings.cv_url
+        settings.save()
+    if opts.only_set_cv_key:
+        print("Key set")  # noqa: T201
         return
 
-    ComicVineTalker.api_key = SETTINGS.cv_api_key
+    ComicVineTalker.api_key = settings.cv_api_key
+    ComicVineTalker.api_base_url = settings.cv_url
 
     signal.signal(signal.SIGINT, signal.SIG_DFL)
 
@@ -138,18 +158,20 @@ def ctmain():
     )
 
     logger.debug("Installed Packages")
-    for pkg in sorted(pkg_resources.working_set, key=lambda x: x.project_name):
-        logger.debug("%s\t%s", pkg.project_name, pkg.version)
+    for pkg in sorted(importlib_metadata.distributions(), key=lambda x: x.name):
+        logger.debug("%s\t%s", pkg.name, pkg.version)
+
+    utils.load_publishers()
+    update_publishers()
 
     if not qt_available and not opts.no_gui:
         opts.no_gui = True
-        print("PyQt5 is not available. ComicTagger is limited to command-line mode.")
-        logger.info("PyQt5 is not available. ComicTagger is limited to command-line mode.")
+        logger.warning("PyQt5 is not available. ComicTagger is limited to command-line mode.")
 
     if opts.no_gui:
         try:
-            cli.cli_mode(opts, SETTINGS)
-        except:
+            cli.cli_mode(opts, settings)
+        except Exception:
             logger.exception("CLI mode failed")
     else:
         os.environ["QtWidgets.QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
@@ -168,12 +190,12 @@ def ctmain():
             import ctypes
 
             myappid = "comictagger"  # arbitrary string
-            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
+            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)  # type: ignore[attr-defined]
             # force close of console window
             swp_hidewindow = 0x0080
-            console_wnd = ctypes.windll.kernel32.GetConsoleWindow()
+            console_wnd = ctypes.windll.kernel32.GetConsoleWindow()  # type: ignore[attr-defined]
             if console_wnd != 0:
-                ctypes.windll.user32.SetWindowPos(console_wnd, None, 0, 0, 0, 0, swp_hidewindow)
+                ctypes.windll.user32.SetWindowPos(console_wnd, None, 0, 0, 0, 0, swp_hidewindow)  # type: ignore[attr-defined]
 
         if platform.system() != "Linux":
             img = QtGui.QPixmap(ComicTaggerSettings.get_graphic("tags.png"))
@@ -181,10 +203,10 @@ def ctmain():
             splash = QtWidgets.QSplashScreen(img)
             splash.show()
             splash.raise_()
-            app.processEvents()
+            QtWidgets.QApplication.processEvents()
 
         try:
-            tagger_window = TaggerWindow(opts.file_list, SETTINGS, opts=opts)
+            tagger_window = TaggerWindow(opts.files, settings, opts=opts)
             tagger_window.setWindowIcon(QtGui.QIcon(ComicTaggerSettings.get_graphic("app.png")))
             tagger_window.show()
 
